//  PyCOM - Distributed Component Model for Python - C++ Client Library
//  Copyright (c) 2011-2012, Dmitry "Divius" Tantsur
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above
//    copyright notice, this list of conditions and the following disclaimer
//    in the documentation and/or other materials provided with the
//    distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <cstring>
#include <iostream>
#include <string>

#include <pycom.h>

#include "pycom-private.h"

namespace pycom {

static const char CALL_COMMAND[4] = {'C', 'A', 'L', 'L'};
static const char OK_RESULT[2] = {'O', 'K'};

Component::Component(_component_pvt *pvt) : _pvt(pvt) {
}

Component::~Component() {
    delete _pvt;
}

Value Component::invoke(const std::string &method_name, const Value &args) {
    Value temp;

    temp["version"] = "1.0";
    temp["interface"] = _pvt->iface_name;
    temp["method"] = method_name;
    temp["args"] = args;
    temp["session_id"] = _pvt->session_id;

    std::string json = _pvt->writer.write(temp);

    zmq::message_t request(sizeof(CALL_COMMAND));
    memcpy(request.data(), CALL_COMMAND, sizeof(CALL_COMMAND));
    _pvt->zsock.send(request, ZMQ_SNDMORE);

    request.rebuild(json.size());
    memcpy(request.data(), json.c_str(), json.size());
    _pvt->zsock.send(request);

    zmq::message_t reply;
    _pvt->zsock.recv(&reply);

    if (reply.size() != sizeof(OK_RESULT) || memcmp(reply.data(), "OK", 2) != 0) {
        std::string error = "pycom: Fatal protocol error";
        std::cerr << error << std::endl;
        throw InternalError(error);
    }

    _pvt->zsock.recv(&reply);

    if (!_pvt->reader.parse((char *)reply.data(), (char *)reply.data() + reply.size(),
            temp, false)) {
        std::string error = "pycom: Error parsing response: " +
            _pvt->reader.getFormattedErrorMessages();
        std::cerr << error << std::endl;
        throw InternalError(error);
    }

    if (temp.isNull()) {
        std::string error = "Empty response";
        std::cerr << error << std::endl;
        throw InternalError(error);
    }

    if (!temp["session_id"].isNull()) {
        _pvt->session_id = temp["session_id"];
    }

    switch (temp["code"].asInt()) {
        case SUCCESS:
            return temp["result"];
        case ERROR_BAD_REQUEST:
            throw BadRequest(temp["error"].asString());
        case ERROR_METHOD_NOT_FOUND:
            throw LookupError(temp["error"].asString());
        default:
            throw RemoteError(temp["code"].asInt(), temp["error"].asString());
    }
}

}
